/*
 * dso_home_wrapper -- a wrapper for bitscope/Dso
 *
 * 2006 (c) jw@suse.de, distribute under GPL v2
 *
 * bitscope/Dso expects its config file BitScope.prb in same directory
 * where the Dso binary is found.
 * The wave-form recorder wants to record its 'WaveData' also there.
 * 
 * This is no good, 
 * a) as multiple user may not want to share their settings.
 * b) would require a user writable installation direcory.
 *
 * We intercept readlink() calls on /proc/self/exe, 
 * so that Dso believes, it is located in $HOME/.bitscope 
 * This has the effect that open() calls for Bitscope.prb and WaveData
 * are made relative to the $HOME/.bitscope directory.
 *
 * We intercept open() calls on the tty-lock file, so that
 * we can run without uucp group-sbit.
 *
 * Intercepting readlink (rather than open) has the advantage that 
 * Path names printed in diagnostics are correct and that future extensions 
 * have a good chance to put thir files in the users home.
 *
 * Usage:
 *
 * gcc -Wall -fPIC -shared -o dso_home_wrapper.so dso_home_wrapper.c -ldl
 * env LD_PRELOAD=`pwd`/dso_home_wrapper.so /bin/ls -l /proc/self/exe
 * env LD_PRELOAD=`pwd`/dso_home_wrapper.so dd if=/etc/passwd of=/var/lock/LCK..ttyUSB0
 */

#include <sys/types.h>	// mode_t
#include <stdlib.h>	// exit
#include <string.h>	// strlen
#include <stdio.h>	// stderr
#define __USE_GNU	// RTLD_NEXT
#include <dlfcn.h>	// dlopen, dlsym

#define PROC_SELF_EXE_PAT	        "/proc/self/exe"
#define PROC_SELF_EXE_FMT	        "%s/.bitscope/Dso"

#define VAR_LOCK_LCK_PRE	       	"/var/lock/LCK.." 
#define VAR_LOCK_LCK_SUBST	       	"/dev/null"

static char buffer[1024];
static int buffer_used;
static int (*real_readlink)();
static int (*real_open)();

static void do_init()
{
  char *dl_error;
  char *home = getenv("HOME");

  dlerror();	/* Clear any existing error */

  // if RTLD_NEXT is unavailable, use handle=dlopen("libc.so.6"); instead.

  real_readlink = (int (*)())dlsym(RTLD_NEXT, "readlink");
  real_open =     (int (*)())dlsym(RTLD_NEXT, "open");
  if ((dl_error = dlerror()) != NULL)
    {
      fprintf(stderr, "dlsym(open): %s\n", dl_error);
      exit(1);
    }

  buffer_used = snprintf(buffer, 1024, PROC_SELF_EXE_FMT, home);
  // printf("readink_wrapper(%s -> %s)\n", PROC_SELF_EXE_PAT, buffer);
}

int readlink(const char *path, char *buf, size_t bufsize)
{
  if (!buffer[0]) do_init();

  if (strcmp(path, PROC_SELF_EXE_PAT))
    return (*real_readlink)(path, buf, bufsize);

  int l = buffer_used;
  if (l > bufsize) l = bufsize;

  fprintf(stderr, "readlink(%s) -> %s\n", path, buffer);

  memcpy((char *)buf, buffer, l);
  return l;
}

int open(const char *path, int mode, int mask)
{
  if (!buffer[0]) do_init();

  if (strncmp(path, VAR_LOCK_LCK_PRE, strlen(VAR_LOCK_LCK_PRE)))
    return (*real_open)(path, mode, mask);

  fprintf(stderr, "open(%s) -> %s\n", path, VAR_LOCK_LCK_SUBST);

  return (*real_open)(VAR_LOCK_LCK_SUBST, mode, mask);
}

int open64(const char *path, int mode, int mask) { return open(path, mode, mask); }
